"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from flask.blueprints import Blueprint
from typing import List, Tuple, Callable
from puffotter.flask.routes.static import define_blueprint as __static
from puffotter.flask.routes.user_management import define_blueprint \
    as __user_management
from puffotter.flask.routes.api.user_management import define_blueprint \
    as __api_user_management

blueprint_generators: List[Tuple[Callable[[str], Blueprint], str]] = [
    (__static, "static"),
    (__user_management, "user_management"),
    (__api_user_management, "api_user_management")
]
"""
Defines the functions used to create the various blueprints
as well as their names
"""



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import os
from puffotter.flask.initialize import init_flask
from puffotter.flask.test.TestFramework import _TestFramework


class TestInitialization(_TestFramework):
    """
    Tests the initialization of the flask application
    """

    def test_missing_environment_variables(self):
        """
        Tests if missing environment variables are detected correctly
        :return: None
        """
        os.environ.pop("FLASK_SECRET")
        try:
            init_flask("puffotter", "", "", self.config, [], [])
            self.fail()
        except SystemExit:
            pass

    def test_missing_required_template(self):
        """
        Tests if missing template files are detected correctly
        :return: None
        """
        for required in self.config.REQUIRED_TEMPLATES.values():
            path = os.path.join(self.temp_templates_dir, required)
            os.remove(path)
            try:
                init_flask("puffotter", "", "", self.config, [], [])
                self.fail()
            except SystemExit:
                with open(path, "w") as f:
                    f.write("")



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Optional, TYPE_CHECKING, List
from puffotter.flask.Config import Config
from puffotter.flask.base import db
from puffotter.flask.db.ModelMixin import ModelMixin
from puffotter.crypto import verify_password
if TYPE_CHECKING:
    from puffotter.flask.db.ApiKey import ApiKey
    from puffotter.flask.db.TelegramChatId import TelegramChatId


class User(ModelMixin, db.Model):
    """
    Model that describes the 'users' SQL table
    A User stores a user's information, including their email address, username
    and password hash
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the Model
        :param args: The constructor arguments
        :param kwargs: The constructor keyword arguments
        """
        super().__init__(*args, **kwargs)

    __tablename__ = "users"
    """
    The name of the table
    """

    username: str = db.Column(
        db.String(Config.MAX_USERNAME_LENGTH),
        nullable=False,
        unique=True
    )
    """
    The user's username
    """

    email: str = db.Column(db.String(150), nullable=False, unique=True)
    """
    The user's email address
    """

    password_hash: str = db.Column(db.String(255), nullable=False)
    """
    The user's hashed password, salted and hashed.
    """

    confirmed: bool = db.Column(db.Boolean, nullable=False, default=False)
    """
    The account's confirmation status. Logins should be impossible as long as
    this value is False.
    """

    confirmation_hash: str = db.Column(db.String(255), nullable=False)
    """
    The account's confirmation hash. This is the hash of a key emailed to
    the user. Only once the user follows the link in the email containing the
    key will their account be activated
    """

    telegram_chat_id: Optional["TelegramChatId"] = db.relationship(
        "TelegramChatId",
        uselist=False,
        back_populates="user",
        cascade="all, delete"
    )
    """
    Telegram chat ID for the user if set up
    """

    api_keys: List["ApiKey"] = db.relationship(
        "ApiKey", back_populates="user", cascade="all, delete"
    )
    """
    API keys for this user
    """

    @property
    def is_authenticated(self) -> bool:
        """
        Property required by flask-login
        :return: True if the user is confirmed, False otherwise
        """
        return True

    @property
    def is_anonymous(self) -> bool:
        """
        Property required by flask-login
        :return: True if the user is not confirmed, False otherwise
        """
        return not self.is_authenticated  # pragma: no cover

    @property
    def is_active(self) -> bool:
        """
        Property required by flask-login
        :return: True
        """
        return self.confirmed

    def get_id(self) -> str:
        """
        Method required by flask-login
        :return: The user's ID as a unicode string
        """
        return str(self.id)

    def verify_password(self, password: str) -> bool:
        """
        Verifies a password against the password hash
        :param password: The password to check
        :return: True if the password matches, False otherwise
        """
        return verify_password(password, self.password_hash)

    def verify_confirmation(self, confirmation_key: str) -> bool:
        """
        Verifies a confirmation key against the confirmation hash
        :param confirmation_key: The key to check
        :return: True if the key matches, False otherwise
        """
        return verify_password(confirmation_key, self.confirmation_hash)



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""


from enum import Enum


class AlertSeverity(Enum):
    """
    Enumeration that defines the various levels of severity an alert can have
    """

    SUCCESS = "success"
    """
    Translates to a green alert
    """

    INFO = "info"
    """
    Translates to a blue alert
    """

    WARNING = "warning"
    """
    Translates to a yellow alert
    """

    DANGER = "danger"
    """
    Translates to a red alert
    """



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""


def byte_string_to_byte_count(byte_string: str) -> int:
    """
    Converts a string representing bytes to a number of bytes.
    For example: "500K" -> 500 000
                 "2.5M" -> 2 500 000
                 "10GB" -> 10 000 000 000
                 "30kb/s" -> 30 000
    :param byte_string: The string to convert
    :return: The amount of bytes
    """
    byte_string = byte_string.lower()

    units = {
        "k": 1000,
        "m": 1000000,
        "g": 1000000000,
        "t": 1000000000000,
        "p": 1000000000000000,
        "e": 1000000000000000000
    }

    for unit in units:
        byte_string = byte_string.replace(unit + "b/s", unit)
        byte_string = byte_string.replace(unit + "b", unit)

    multiplier = 1
    byte_num = ""
    for i, char in enumerate(byte_string):
        if char.isdigit():
            byte_num += char
        else:
            # Unit should be last symbol in string
            if len(byte_string) - 1 != i:
                raise ValueError()
            else:
                try:
                    multiplier = units[char]
                except KeyError:
                    raise ValueError()

    return multiplier * int(byte_num)


def human_readable_bytes(
        bytecount: int,
        base_1024: bool = False,
        remove_trailing_zeroes: bool = True
) -> str:
    """
    Converts an amount of bytes into a human-readable string
    :param bytecount: The bytes to convert
    :param base_1024: Whether or not to use 1024 as base (for mebibytes etc)
    :param remove_trailing_zeroes: If set to True, will remove any trailing
                                   zeroes from the string
    :return: The human-readable string
    """
    units = ["K", "M", "G", "T", "P", "E", "Z", "Y"]
    unit_index = -1
    base = 1024 if base_1024 else 1000
    _bytes = float(bytecount)

    while True:
        _bytes /= base
        unit_index += 1

        if int(_bytes) < base or unit_index == len(units) - 1:
            break

    # Formatting
    bytestring = ("%.3f" % _bytes)

    if remove_trailing_zeroes:
        string_index = len(bytestring) - 1
        while bytestring[string_index] == "0":
            string_index -= 1
        bytestring = bytestring[0:string_index + 1]

        if bytestring.endswith("."):
            bytestring = bytestring[0:-1]

    i = "i" if base_1024 else ""
    return bytestring + units[unit_index] + i + "B"



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of fat-ffipd.

fat-ffipd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

fat-ffipd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with fat-ffipd.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Any
from flask import Blueprint, request
from flask_login import login_required
from puffotter.crypto import generate_random, generate_hash
from puffotter.flask.base import db
from puffotter.flask.Config import Config
from puffotter.flask.db.User import User
from puffotter.flask.db.ApiKey import ApiKey
from puffotter.flask.exceptions import ApiException
from puffotter.flask.routes.decorators import api, api_login_required


def define_blueprint(blueprint_name: str) -> Blueprint:
    """
    Defines the blueprint for this route
    :param blueprint_name: The name of the blueprint
    :return: The blueprint
    """
    blueprint = Blueprint(blueprint_name, __name__)
    api_base = f"/api/v{Config.API_VERSION}"

    @blueprint.route(f"{api_base}/key", methods=["POST", "DELETE"])
    @api
    def api_key() -> Dict[str, Any]:
        """
        Allows users to request a new API key or revoke an existing API key
        :return: The JSON response
        """
        data = request.get_json()
        if request.method == "POST":
            username = data["username"]
            password = data["password"]
            user = User.query.filter_by(username=username).first()

            if user is None:
                raise ApiException("user does not exist", 401)
            elif not user.confirmed:
                raise ApiException("user is not confirmed", 401)
            elif not user.verify_password(password):
                raise ApiException("password is incorrect", 401)
            else:
                key = generate_random(32)
                hashed = generate_hash(key)
                _api_key = ApiKey(user=user, key_hash=hashed)
                db.session.add(_api_key)
                db.session.commit()

                return {
                    "api_key": "{}:{}".format(_api_key.id, key),
                    "expiration": (
                            int(_api_key.creation_time)
                            + Config.MAX_API_KEY_AGE
                    ),
                    "user": user.__json__(True)
                }

        else:  # request.method == "DELETE"
            key = data["api_key"]
            _api_key = ApiKey.query.get(key.split(":", 1)[0])

            if _api_key is None:
                raise ApiException("api key does not exist", 401)
            elif not _api_key.verify_key(key):
                raise ApiException("api key not valid", 401)
            else:
                db.session.delete(_api_key)
                db.session.commit()
                return {}

    @blueprint.route(f"{api_base}/authorize", methods=["GET"])
    @api_login_required
    @login_required
    @api
    def api_authorize() -> Dict[str, Any]:
        """
        Allows a user to check if an API key is authorized or not
        :return: None
        """
        return {}  # Checks done by @login_required

    return blueprint
